# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/processors/02_eoq_preprocessors.ipynb.

# %% auto 0
__all__ = ['GetInventory']

# %% ../../../nbs/agents/processors/02_eoq_preprocessors.ipynb 4
# General libraries:
import numpy as np

# %% ../../../nbs/agents/processors/02_eoq_preprocessors.ipynb 6
class GetInventory():
    """
    This class implements a processor that get the inventory from the state.

    Args:
        num_features (int): Number of features in the state.
        num_products (int): Number of products in the state.

    """
    

    def __init__(self, num_features, num_products, max_l = None):
        self.num_features = num_features
        self.num_products = num_products
        self.max_l = max_l
    
    def __call__(self, input):
        #assert len(input) == self.num_features + self.num_products
        assert len(input.shape) == 1
        
        if self.max_l is not None:
            return input[self.num_features:self.num_features+self.num_products+np.max(self.max_l)*self.num_products]
        else:
            return input[self.num_features:self.num_features+self.num_products]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/rl_agents/01_SAC_gumbel.ipynb.

# %% auto 0
__all__ = ['GumbelSoftmax', 'SAC_gumbelPolicy', 'SAC_gumbel']

# %% ../../../nbs/agents/rl_agents/01_SAC_gumbel.ipynb 4
# General libraries
import numpy as np
from copy import deepcopy

# PyTorch
import torch
import torch.optim as optim
import torch.nn.functional as F

# Mushroom
from mushroom_rl.utils.parameters import Parameter
from mushroom_rl.policy import Policy
from mushroom_rl.algorithms.actor_critic import DeepAC
from mushroom_rl.utils.parameters import to_parameter
from mushroom_rl.utils.replay_memory import ReplayMemory
from mushroom_rl.approximators import Regressor
from mushroom_rl.approximators.parametric.torch_approximator import *

# %% ../../../nbs/agents/rl_agents/01_SAC_gumbel.ipynb 6
class GumbelSoftmax(torch.distributions.RelaxedOneHotCategorical):
    
    '''
    A differentiable Categorical distribution using reparametrization trick with Gumbel-Softmax
    Explanation http://amid.fish/assets/gumbel.html
    NOTE: use this in place PyTorch's RelaxedOneHotCategorical distribution since its log_prob is not working right (returns positive values)
    Papers:
    [1] The Concrete Distribution: A Continuous Relaxation of Discrete Random Variables (Maddison et al, 2017)
    [2] Categorical Reparametrization with Gumbel-Softmax (Jang et al, 2017)
    '''

    def sample(self, sample_shape=torch.Size()):
        '''Gumbel-softmax sampling. Note rsample is inherited from RelaxedOneHotCategorical'''
        u = torch.empty(self.logits.size(), device=self.logits.device, dtype=self.logits.dtype).uniform_(0, 1)
        noisy_logits = self.logits - torch.log(-torch.log(u))
        return torch.argmax(noisy_logits, dim=-1)

    def rsample(self, sample_shape=torch.Size()):
        '''
        Gumbel-softmax resampling using the Straight-Through trick.
        To see standalone code of how this works, refer to https://gist.github.com/yzh119/fd2146d2aeb329d067568a493b20172f
        '''
        rout = super().rsample(sample_shape)  # differentiable
        out = F.one_hot(torch.argmax(rout, dim=-1), self.logits.shape[-1]).float()
        return (out - rout).detach() + rout

    def log_prob(self, value):
        '''value is one-hot or relaxed'''
        if value.shape != self.logits.shape:
            value = F.one_hot(value.long(), self.logits.shape[-1]).float()
            assert value.shape == self.logits.shape
        return - torch.sum(- value * F.log_softmax(self.logits, -1), -1)
    
    def entropy(self):
        min_real = torch.finfo(self.logits.dtype).min
        logits = torch.clamp(self.logits, min=min_real)
        p_log_p = logits * self.probs
        return -p_log_p.sum(-1)

# %% ../../../nbs/agents/rl_agents/01_SAC_gumbel.ipynb 8
class SAC_gumbelPolicy(Policy):
    """
    Class used to implement the policy used by the Soft Actor-Critic
    algorithm. The policy is a relaxed categorcial distribution policy.
    This class implements the compute_action_and_log_prob and the
    compute_action_and_log_prob_t methods, that are fundamental for
    the internals calculations of the SAC algorithm.
    """
    def __init__(self, actor_approximator, temperature=1.0):
        """
        Constructor.
        Args:
            actor_approximator (Regressor): a regressor computing categorical action logits given a
                state;
        """
        self._actor_approximator = actor_approximator
        self._temperature = torch.tensor(temperature)

        use_cuda = self._actor_approximator.model.use_cuda

        # if use_cuda:
        #     self._delta_a = self._delta_a.cuda()
        #     self._central_a = self._central_a.cuda()

        self._add_save_attr(
            _actor_approximator='mushroom',
            _temperature='torch',
        )

    def __call__(self, state, action):
        raise NotImplementedError

    def draw_action(self, state):        
        action = self.compute_action_and_log_prob_t(state, compute_log_prob=False).detach().cpu().numpy()
        return action.astype('int') # So that gym recognizes it as one hot integers

    def compute_action_and_log_prob(self, state):
        """
        Function that samples actions using the reparametrization trick and
        the log probability for such actions.
        Args:
            state (np.ndarray): the state in which the action is sampled.
        Returns:
            The actions sampled and the log probability as numpy arrays.
        """
        a, log_prob = self.compute_action_and_log_prob_t(state)
        return a.detach().cpu().numpy(), log_prob.detach().cpu().numpy()

    def compute_action_and_log_prob_t(self, state, compute_log_prob=True):
        """
        Function that samples actions using the reparametrization trick and,
        optionally, the log probability for such actions.
        Args:
            state (np.ndarray): the state in which the action is sampled;
            compute_log_prob (bool, True): whether to compute the log
            probability or not.
        Returns:
            The actions sampled and, optionally, the log probability as torch
            tensors.
        """
        dist = self.distribution(state)
        a = dist.rsample()

        if compute_log_prob:
            log_prob = dist.log_prob(a)
            return a, log_prob
        else:
            return a

    def distribution(self, state):
        """
        Compute the policy distribution in the given states.
        Args:
            state (np.ndarray): the set of states where the distribution is
                computed.
        Returns:
            The torch distribution for the provided states.
        """
        logits = self._actor_approximator.predict(state, output_tensor=True)

        return GumbelSoftmax(temperature=self._temperature, logits=logits)

    def entropy(self, state=None):
        """
        Compute the entropy of the policy.
        Args:
            state (np.ndarray): the set of states to consider.
        Returns:
            The value of the entropy of the policy.
        """
        return torch.mean(self.distribution(state).entropy()).detach().cpu().numpy().item()

    def reset(self):
        pass

    def set_weights(self, weights):
        """
        Setter.
        Args:
            weights (np.ndarray): the vector of the new weights to be used by
                the policy.
        """

        self._actor_approximator.set_weights(weights)

    def get_weights(self):
        """
        Getter.
        Returns:
             The current policy weights.
        """

        return self._actor_approximator.get_weights()

    @property
    def use_cuda(self):
        """
        True if the policy is using cuda_tensors.
        """
        return self._actor_approximator.model.use_cuda

    def parameters(self):
        """
        Returns the trainable policy parameters, as expected by torch
        optimizers.
        Returns:
            List of parameters to be optimized.
        """
        return self._actor_approximator.model.network.parameters()

# %% ../../../nbs/agents/rl_agents/01_SAC_gumbel.ipynb 10
class SAC_gumbel(DeepAC):
    """
    Soft Actor-Critic algorithm.
    "Soft Actor-Critic Algorithms and Applications".
    Haarnoja T. et al.. 2019.
    """
    def __init__(self, mdp_info, actor_params,
                 actor_optimizer, critic_params, batch_size,
                 initial_replay_size, max_replay_size, warmup_transitions, tau,
                 lr_alpha, temperature=1.0, target_entropy=None,
                 critic_fit_params=None):
        """
        Constructor.
        Args:
            actor_params (dict): parameters of the actor approximator
                to build;
            actor_optimizer (dict): parameters to specify the actor
                optimizer algorithm;
            critic_params (dict): parameters of the critic approximator to
                build;
            batch_size ((int, Parameter)): the number of samples in a batch;
            initial_replay_size (int): the number of samples to collect before
                starting the learning;
            max_replay_size (int): the maximum number of samples in the replay
                memory;
            warmup_transitions ([int, Parameter]): number of samples to accumulate in the
                replay memory to start the policy fitting;
            tau ([float, Parameter]): value of coefficient for soft updates;
            lr_alpha ([float, Parameter]): Learning rate for the entropy coefficient;
            temperature (float): the temperature for the softmax part of the gumbel reparametrization
            target_entropy (float, None): target entropy for the policy, if
                None a default value is computed ;
            critic_fit_params (dict, None): parameters of the fitting algorithm
                of the critic approximator.
        """
        self._critic_fit_params = dict() if critic_fit_params is None else critic_fit_params

        self._batch_size = to_parameter(batch_size)
        self._warmup_transitions = to_parameter(warmup_transitions)
        self._tau = to_parameter(tau)

        if target_entropy is None:
            self._target_entropy = -np.prod(mdp_info.action_space.n).astype(np.float32)
        else:
            self._target_entropy = target_entropy

        self._replay_memory = ReplayMemory(initial_replay_size, max_replay_size)

        if 'n_models' in critic_params.keys():
            assert critic_params['n_models'] == 2
        else:
            critic_params['n_models'] = 2

        target_critic_params = deepcopy(critic_params)
        self._critic_approximator = Regressor(TorchApproximator,
                                              **critic_params)
        self._target_critic_approximator = Regressor(TorchApproximator,
                                                     **target_critic_params)

        actor_approximator = Regressor(TorchApproximator,
                                          **actor_params)

        policy = SAC_gumbelPolicy(actor_approximator, temperature)

        self._init_target(self._critic_approximator,
                          self._target_critic_approximator)

        self._log_alpha = torch.tensor(0., dtype=torch.float32)

        if policy.use_cuda:
            self._log_alpha = self._log_alpha.cuda().requires_grad_()
        else:
            self._log_alpha.requires_grad_()

        self._alpha_optim = optim.Adam([self._log_alpha], lr=lr_alpha)

        policy_parameters = actor_approximator.model.network.parameters()

        self._add_save_attr(
            _critic_fit_params='pickle',
            _batch_size='mushroom',
            _warmup_transitions='mushroom',
            _tau='mushroom',
            _target_entropy='primitive',
            _replay_memory='mushroom',
            _critic_approximator='mushroom',
            _target_critic_approximator='mushroom',
            _log_alpha='torch',
            _alpha_optim='torch'
        )

        super().__init__(mdp_info, policy, actor_optimizer, policy_parameters)

    def fit(self, dataset):
        self._replay_memory.add(dataset)
        if self._replay_memory.initialized:
            state, action, reward, next_state, absorbing, _ = \
                self._replay_memory.get(self._batch_size())

            if self._replay_memory.size > self._warmup_transitions():
                action_new, log_prob = self.policy.compute_action_and_log_prob_t(state)
                loss = self._loss(state, action_new, log_prob)
                self._optimize_actor_parameters(loss)
                self._update_alpha(log_prob.detach())

            q_next = self._next_q(next_state, absorbing)
            q = reward + self.mdp_info.gamma * q_next

            self._critic_approximator.fit(state, action, q,
                                          **self._critic_fit_params)

            self._update_target(self._critic_approximator,
                                self._target_critic_approximator)

    def _loss(self, state, action_new, log_prob):
        q_0 = self._critic_approximator(state, action_new,
                                        output_tensor=True, idx=0)
        q_1 = self._critic_approximator(state, action_new,
                                        output_tensor=True, idx=1)

        q = torch.min(q_0, q_1)

        return (self._alpha * log_prob - q).mean()

    def _update_alpha(self, log_prob):
        alpha_loss = - (self._log_alpha * (log_prob + self._target_entropy)).mean()
        self._alpha_optim.zero_grad()
        alpha_loss.backward()
        self._alpha_optim.step()

    def _next_q(self, next_state, absorbing):
        """
        Args:
            next_state (np.ndarray): the states where next action has to be
                evaluated;
            absorbing (np.ndarray): the absorbing flag for the states in
                ``next_state``.
        Returns:
            Action-values returned by the critic for ``next_state`` and the
            action returned by the actor.
        """
        a, log_prob_next = self.policy.compute_action_and_log_prob(next_state)

        q = self._target_critic_approximator.predict(
            next_state, a, prediction='min') - self._alpha_np * log_prob_next
        q *= 1 - absorbing

        return q

    def _post_load(self):
        self._update_optimizer_parameters(self.policy.parameters())

    @property
    def _alpha(self):
        return self._log_alpha.exp()

    @property
    def _alpha_np(self):
        return self._alpha.detach().cpu().numpy()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/benchmark_agents/12_SS.ipynb.

# %% auto 0
__all__ = ['SSAgent', 'SSPolicy']

# %% ../../../nbs/agents/benchmark_agents/12_SS.ipynb 4
# General libraries:
import numpy as np
from scipy.stats import norm
from tqdm import tqdm

# Mushroom libraries
from mushroom_rl.core import Agent

# %% ../../../nbs/agents/benchmark_agents/12_SS.ipynb 6
class SSAgent(Agent):

    """
    Agent implementing the QR policy.

    # TODO adjust description

    Args:
        mdp_info (MDPInfo): Information about the Markov Decision Process (MDP).
        s (numpy.ndarray): The fixed ordering cost.
        h (numpy.ndarray): The holding cost per unit per period.
        l (numpy.ndarray): The lead time per product.
        preprocessors (list): List of preprocessors to be applied to the state.
        postprocessors (list): List of postprocessors to be applied to the policy.
        agent_name (str): Name of the agent. If set to None will use some default name.
        precision (int): Number of decimal places to round the demand input to.

    Attributes:
        mdp_info (MDPInfo): Information about the Markov Decision Process (MDP).
        policy (EOQPolicy): The EOQ policy implemented by the agent.

    """


    def __init__(self,
                  mdp_info,
                  mdp,
                  s, # fixed ordering cost
                  h, # holding cost per unit per period
                  l, # lead time
                  p, # penalty cost per unit
                  unit_size = 0.01,
                  preprocessors = None,
                  postprocessors = None,
                  agent_name = None,
                  precision = 5,
                  num_iterations_per_parameter = 12,
                  include_l = True
        ):

        mdp._mdp_info.horizon = mdp.demand.shape[0]
        mdp.reset(0)

        policy = SSPolicy(
            d = None,
            s = s,
            h = h,
            l = l,
            p = p,
            mdp = mdp,
            unit_size = unit_size,
            preprocessors = preprocessors,
            postprocessors = postprocessors,
            num_iterations_per_parameter = num_iterations_per_parameter,
            include_l = include_l,
        )

        self.precision=precision

        if agent_name is None:
            self.name = 'SSAgent'
        else:
            self.name = agent_name
        
        self.train_directly=True
        self.train_mode = "direct"

        super().__init__(mdp_info, policy)


    def fit(self, features = None, demand=None):

        """ 
        Fit the QR policy to the given demand.

        # TODO adjust description

        This method allows the EOQ agent to adapt its policy to historic demand data, assuming a fixed demand rate without uncertainty.

        Parameters:
            demand (numpy.ndarray): The demand for each period and product. The array should have the shape (num_products, num_periods).

        Returns:
            None
        """

        assert isinstance(demand, np.ndarray)
        assert demand.ndim == 2

        self.policy.set_s_S(demand, self._preprocessors[0])

class SSPolicy():

    """
    Policy implementing the QR strategy.

    Notethat d, s, and h must all have the shape (num_products,)

    # TODO adjust description

    Args:
        d (numpy.ndarray): The (average) demand per period for each product.
        s (numpy.ndarray): The fixed ordering cost for each product.
        h (numpy.ndarray): The holding cost per unit per period for each product.
        l (numpy.ndarray): The lead time per product.
        postprocessors (list): List of postprocessors to be applied to the action.

    Attributes:
        d (numpy.ndarray): The (average) demand per period for each product.
        s (numpy.ndarray): The fixed ordering cost for each product.
        h (numpy^.ndarray): The holding cost per unit per period for each product.
        l (numpy.ndarray): The lead time per product.
        num_products (int): The number of products.
        q_star (numpy.ndarray): The optimal order quantity per product.
        postprocessors (list): List of postprocessors to be applied to the action.

    """

    def __init__(self,
                 d, 
                 s, 
                 h, 
                 l, 
                 p,
                 mdp,
                 unit_size = 0.01,
                 preprocessors = None,
                 postprocessors = None,
                 num_iterations_per_parameter = 12,
                 include_l = True
                 ):
        self.d = d
        self.s = s
        self.h = h
        self.l = l
        self.p = p
        self.unit_size = unit_size
        self.num_products = len(s)

        if preprocessors is None:
            self.preprocessors = []
        else:
            self.preprocessors = (preprocessors)
        if postprocessors is None:
            self.postprocessors = []
        else:
            self.postprocessors = (postprocessors)

        self.mdp = mdp

        self.num_iterations_per_parameter = num_iterations_per_parameter

        self.include_l = include_l

        self.s = 0.5 # initial value
        self.S = 0.5 # initial value
    
    def calculate_grid(self, s, S, step_size, plot=False):

        s_candidates = np.arange(s-step_size*self.num_iterations_per_parameter, s+step_size*self.num_iterations_per_parameter, step_size)
        S_candidates = np.arange(S-step_size*self.num_iterations_per_parameter, S+step_size*self.num_iterations_per_parameter, step_size)

        # only use positive or zero candidates
        s_candidates = s_candidates[s_candidates >= -0.1]
        S_candidates = S_candidates[S_candidates >= 0]

        cost_matrix = np.zeros((len(s_candidates), len(S_candidates)))

        for i, s_candidate in enumerate(s_candidates):
            for j, S_candidate in enumerate(S_candidates):
                cost_value = self.run_simulation(s_candidate, S_candidate)

                cost_matrix[i, j] = cost_value
                print("s:", np.round(s_candidate,2), "S:", np.round(S_candidate,2), "cost:", np.round(cost_value))

        best_cost = np.min(cost_matrix)
        best_s_index, best_S_index = np.unravel_index(np.argmin(cost_matrix), cost_matrix.shape)
        best_s = s_candidates[best_s_index]
        best_S = S_candidates[best_S_index]

        return best_s, best_S, best_cost
    
    def run_simulation(self, s, S):

        total_cost = 0

        state = self.mdp.reset(0)

        for t in range(self.mdp.info.horizon):

            state = self.preprocessor(state)
            action = self.draw_action_train(state, s, S)
                
            state, reward, _, _ = self.mdp.step(action)
            
            total_cost += -reward

        return total_cost

    def set_s_S(self, demand, preprocessor):
        
        """
        Set the optimal order quantity (q_star) for each product.

        This method calculates and assigns the optimal order quantity based on the EOQ formula.

        Returns:
            None

        """

        self.preprocessor = preprocessor

        self.s, self.S, cost = self.calculate_grid(self.s, self.S, step_size = self.unit_size*10)
        print("iteration 1:", "s:", self.s, "S:", self.S, "cost:", cost)
        self.s, self.S, cost = self.calculate_grid(self.s, self.S, step_size = self.unit_size*3)
        print("iteration 2:", "s:", self.s, "S:", self.S, "cost:", cost)
        self.s, self.S, cost = self.calculate_grid(self.s, self.S, step_size = self.unit_size)
        print("iteration 4:", "s:", self.s, "S:", self.S, "cost:", cost)


    def draw_action(self, input):

        """
        Generate an action based on the current state.

        # TODO adjust description

        Returns zero for products which have still sufficient inventory, and the optimal order quantity for products which are running out of stock.

        Parameters:
            input (numpy.ndarray): The current inventory level and potentially order pipeline for each product.

        Returns:
            numpy.ndarray: The action to be taken, indicating the quantity to order for each product.

        """

        for preprocessor in self.preprocessors:
            input = preprocessor(input)

        if self.include_l:
            total_inventory_position = np.sum(input)
            if total_inventory_position <= self.s:
                q = self.S - total_inventory_position
            else:
                q = 0 
        else:
            print("Only policy based on total inventory position implemented")
        

        action = np.array([q])

        for postprocessor in self.postprocessors:
            action = postprocessor(action)

        return action
    
    def draw_action_train(self, input, s, S):

        """
        Generate an action based on the current state.

        # TODO adjust description

        Returns zero for products which have still sufficient inventory, and the optimal order quantity for products which are running out of stock.

        Parameters:
            input (numpy.ndarray): The current inventory level and potentially order pipeline for each product.

        Returns:
            numpy.ndarray: The action to be taken, indicating the quantity to order for each product.

        """

        #print("R:", R, "Q:", Q)

        for preprocessor in self.preprocessors:
            input = preprocessor(input)

        if self.include_l:
            total_inventory_position = np.sum(input)
            if total_inventory_position <= s:
                q = S - total_inventory_position
            else:
                q = 0 
                
        action = np.array([q])

        for postprocessor in self.postprocessors:
            action = postprocessor(action)

        return action
    
    def reset(self):
        pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/environments/21_feature_converters.ipynb.

# %% auto 0
__all__ = ['feature_oracle', 'get_time_series_features']

# %% ../../nbs/environments/21_feature_converters.ipynb 4
# General libraries:
import numpy as np

# %% ../../nbs/environments/21_feature_converters.ipynb 5
def feature_oracle(features, window = 10, flat = False):
    """

    Takes a matrix of features where n is the number of time steps and m is the number of features.
    For each feature it adds as many columns as the window size where each column is the feature n+1 time steps in the future.

    Args:
        features (np.array): Array of features.
        window (int): Window size how many future features to add
        flat (bool): If True, the output is a 2D array. If False, the output is a 3D array.
    """

    oracle_features = np.zeros((features.shape[0]-window, features.shape[1], window+1))

    for i in range(window+1):
        if i == window:
            oracle_features[:, :, i] = features[i:,:].copy()
        else:
            oracle_features[:, :, i] = features[i:-window+i,:].copy()

    if flat:
        oracle_features = oracle_features.reshape((oracle_features.shape[0], oracle_features.shape[1] * oracle_features.shape[2]))

    return oracle_features

# %% ../../nbs/environments/21_feature_converters.ipynb 6
def get_time_series_features(time_series, window = 10, flat = False, include_t = False):

    # TODO Check how to differentiate between labels and features (features may also have the feature information
    # TODO of the current period, not just the lagged information)

    """

    Takes a matrix of a time series with n time steps and m features. It returns a 3D array where the first dimension
    is the time step, the second dimension is the window size and the third dimension is the number of features.

    Args:
        features (np.array): Input time series.
        window (int): Window size how many past time steps to include
        flat (bool): If True, the output is a 2D array. If False, the output is a 3D array.
    """

    num_features = time_series.shape[1]

    series = np.zeros((len(time_series)-window, window, num_features))

    if include_t:
        for i in range(len(time_series) - window):
            series[i,:,:]=(time_series[i+1:i+window+1])
    else:
        for i in range(len(time_series) - window):
            series[i,:,:]=(time_series[i:i+window])

    if flat:
        series = series.reshape((series.shape[0], series.shape[1] * series.shape[2]))

    return series



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/benchmark_agents/21_Capped_base_stock.ipynb.

# %% auto 0
__all__ = ['CBSAgent', 'CBSPolicy']

# %% ../../../nbs/agents/benchmark_agents/21_Capped_base_stock.ipynb 4
# General libraries:
import numpy as np
from scipy.stats import norm
from tqdm import tqdm

# Mushroom libraries
from mushroom_rl.core import Agent

# %% ../../../nbs/agents/benchmark_agents/21_Capped_base_stock.ipynb 6
class CBSAgent(Agent):

    train_directly=True

    """_summary_
    """

    def __init__(self,
                    mdp_info,
                    mdp,
                    h, # holding cost per unit per period
                    p, # penalty cost per unit
                    l, # lead time
                    unit_size = 0.01,
                    preprocessors = None,
                    postprocessors = None,
                    agent_name = None,
                    precision=5,
                    num_iterations_per_parameter=12,
                    include_l=True
                    ):
        mdp._mdp_info.horizon = mdp.demand.shape[0]
        mdp.reset(0)

        policy = CBSPolicy(
                        d=None, # Demand
                        h=h, # holding cost per unit per period
                        p=p, # penalty cost per unit
                        l=l, # lead time
                        mdp = mdp,
                        unit_size = unit_size,
                        preprocessors = preprocessors,
                        postprocessors = postprocessors,
                        num_iterations_per_parameter=num_iterations_per_parameter,
                        include_l=include_l
                        )
        self.precision=precision

        if agent_name is None:
            self.name = 'CBSAgent'
        else:
            self.name = agent_name
        
        self.train_directly=True
        self.train_mode="direct"

        super().__init__(mdp_info, policy)
    
    def fit(self, features=None, demand=None):
        assert isinstance(demand, np.ndarray)
        assert demand.ndim == 2

        self.policy.set_params(demand, self._preprocessors[0])

class CBSPolicy():
    """
    

    """

    def __init__(self,
                    d, 
                    h,
                    p,
                    l,
                    mdp,
                    unit_size,
                    preprocessors,
                    postprocessors,
                    num_iterations_per_parameter,
                    include_l
                    ):

        self.d = d
        self.h = h
        self.p = p
        self.l = l
        self.unit_size = unit_size

        # num products len(h) if h is not scalar, otherwise 1
        self.num_products = len(h) if isinstance(h, np.ndarray) else 1

        if preprocessors is None:
            self.preprocessors = list()
        else:
            self.preoprocessors = (preprocessors)
        if postprocessors is None:
            self.postprocessors = list()
        else:
            self.postprocessors = (postprocessors)

        self.mdp=mdp
        self.num_iterations_per_parameter=num_iterations_per_parameter
        self.include_l=include_l

        self.base = 0.5
        self.cap = 0.5
    

    def calculate_grid(self, base, cap, step_size, plot=False):

        base_candidates = np.arange(base-step_size*self.num_iterations_per_parameter, base+step_size*self.num_iterations_per_parameter, step_size)
        cap_candidates = np.arange(cap-step_size*self.num_iterations_per_parameter, cap+step_size*self.num_iterations_per_parameter, step_size)

        # only use positive or zero candidates
        base_candidates = base_candidates[base_candidates >= -0.1]
        cap_candidates = cap_candidates[cap_candidates >= -0.1]

        cost_matrix = np.zeros((len(base_candidates), len(cap_candidates)))

        for i, base_candidate in enumerate(base_candidates):
            for j, cap_candidate in enumerate(cap_candidates):
                cost_value = self.run_simulation(base_candidate, cap_candidate)

                cost_matrix[i, j] = cost_value
                print("base:", np.round(base_candidate,2), "cap:", np.round(cap_candidate,2), "cost:", np.round(cost_value))

        best_cost = np.min(cost_matrix)
        best_base_index, best_cap_index = np.unravel_index(np.argmin(cost_matrix), cost_matrix.shape)

        best_base = base_candidates[best_base_index]
        best_cap = cap_candidates[best_cap_index]

        return best_base, best_cap, best_cost

    def run_simulation(self, base, cap):

        total_cost = 0
        state = self.mdp.reset(0)

        for t in range(self.mdp.info.horizon):

            state = self.preprocessor(state)
            action = self.draw_action_train(state, base, cap)

            state, reward, _, _ = self.mdp.step(action)

            total_cost += -reward

        return total_cost
    
    def set_params(self, demand, preprocessor):
        
        self.preprocessor = preprocessor

        self.base, self.cap, cost = self.calculate_grid(self.base, self.cap, step_size = self.unit_size*10)
        print("iteration 1:", "base:", self.base, "cap:", self.cap, "cost:", cost)
        self.base, self.cap, cost = self.calculate_grid(self.base, self.cap, step_size = self.unit_size*3)
        print("iteration 2:", "base:", self.base, "cap:", self.cap, "cost:", cost)
        self.base, self.cap, cost = self.calculate_grid(self.base, self.cap, step_size = self.unit_size)
        print("iteration 4:", "base:", self.base, "cap:", self.cap, "cost:", cost)

    def draw_action(self, input):

        """
        Note: only designed for single product case
        """

        for preprocessor in self.preprocessors:
            # input must be a vector containing the inventory and the pipeline vector
            input = preprocessor(input)
        
        # print(input)
        # breakpoint()
        
        total_inventory_position = np.sum(input)
        q = np.minimum(np.maximum(self.base - total_inventory_position, 0), self.cap)

        print("action:", q)

        return np.array([q])

    def draw_action_train(self, input, base, cap):

        """
        Note: only designed for single product case
        """

        for preprocessor in self.preprocessors:
            # input must be a vector containing the inventory and the pipeline vector
            input = preprocessor(input)
        
        total_inventory_position = np.sum(input)
        q = np.minimum(np.maximum(base - total_inventory_position, 0), cap)

        return np.array([q])

    def reset(self):
        pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/agents/rl_agents/pre_specified_agents/12_SAC_RNN.ipynb.

# %% auto 0
__all__ = ['SACRNN', 'SACRNN_hybrid']

# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/12_SAC_RNN.ipynb 5
# General libraries
import numpy as np

# Torch
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

# Networks
from ...networks.critics import CriticNetworkStateActionRNN
from ...networks.actors import ActorNetworkRNN_MLP_hybrid, ActorNetworkRNN 
from ...processors.processors import GetTimeSeriesAndStaticFeatures

# Algorithms
from mushroom_rl.algorithms.actor_critic.deep_actor_critic import SAC

# Processors
from ...processors.processors import HybridToContinuous


# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/12_SAC_RNN.ipynb 6
class SACRNN():

    """
    # TODO Update docstring
    Soft Actor Critic (SAC) agent with hybrid action, both based on Gaussian. The binary action is 
    0 if the output of the network is less or equal than 0, and 1 otherwise.

    Args:
        mdp_info (MDPInfo): Contains relevant information about the environment.
        learning_rate_actor (float): Learning rate for the actor.
        learning_rate_critic (float): Learning rate for the critic.
        learning_rate_alpha (float): Learning rate for the temperature parameter.
        initial_replay_size (int): Number of transitions to save in the replay buffer during startup.
        max_replay_size (int): Maximum number of transitions to save in the replay buffer.
        batch_size (int): Number of transitions to sample each time experience is replayed.
        n_features (int): Number of features for the hidden layers of the networks.
        warmup_transitions (int): Number of transitions to replay before starting to update the networks.
        lr_alpha (float): Learning rate for the temperature parameter.
        tau (float): Parameter for the soft update of the target networks.
        optimizer (torch.optim): Optimizer to use for the networks.
        squeeze_output (bool): Whether to squeeze the output of the actor network or not.
        use_cuda (bool): Whether to use CUDA or not. If True and not available, it will use CPU.
        agent_name (str): Name of the agent. If set to None will use some default name.

    """

    def __init__(
            self,
            mdp_info,
            num_time_series_features,
            num_non_time_series_features,
            window_size,
            learning_rate_actor = 3e-4,
            learning_rate_critic = None,
            initial_replay_size = 64,
            max_replay_size = 50000,
            batch_size = 64,
            n_features = 64,
            warmup_transitions = 100,
            lr_alpha = 3e-4,
            tau = 0.005,
            log_std_min = -20,
            log_std_max = 2,
            target_entropy = None,
            RNN_unit = "GRU", # TODO: Make RNN and LSTM possible
            optimizer = optim.Adam,
            squeeze_output = True,
            use_cuda = True,
            agent_name = None):
        
        use_cuda = use_cuda and torch.cuda.is_available()

        input_shape_data = mdp_info.observation_space.shape
        input_shape_RNN = (num_time_series_features, window_size)
        input_shape_MLP = (num_non_time_series_features,)

        actor_output_shape = (mdp_info.action_space.shape[0],)

        if learning_rate_critic is None:
            learning_rate_critic = learning_rate_actor


        actor_mu_params = dict(network=ActorNetworkRNN,
                                n_features=n_features,
                                input_shape=input_shape_data,
                                input_shape_RNN=input_shape_RNN,
                                input_shape_MLP=input_shape_MLP,
                                output_shape=actor_output_shape,
                                input_processor=GetTimeSeriesAndStaticFeatures(num_time_series_features,window_size),
                                use_cuda=use_cuda)

        actor_sigma_params = dict(network=ActorNetworkRNN,
                                    n_features=n_features,
                                    input_shape= input_shape_data,
                                    input_shape_RNN=input_shape_RNN,
                                    input_shape_MLP=input_shape_MLP,
                                    output_shape=actor_output_shape,
                                    input_processor=GetTimeSeriesAndStaticFeatures(num_time_series_features,window_size),
                                    use_cuda=use_cuda)
        
        actor_optimizer = {'class': optimizer,
                    'params': {'lr': learning_rate_actor}} 
        
        critic_input_shape = (input_shape_data[0] + actor_output_shape[0],)
        input_shape_RNN = (num_time_series_features, window_size)
        input_shape_MLP = (num_non_time_series_features + actor_output_shape[0],)

        critic_params = dict(network=CriticNetworkStateActionRNN,
                        optimizer={'class': optim.Adam,
                                'params': {'lr': learning_rate_critic}}, 
                        loss=F.mse_loss,
                        n_features=n_features,
                        input_shape=critic_input_shape,
                        input_shape_RNN = input_shape_RNN,
                        input_shape_MLP = input_shape_MLP,
                        output_shape=(1,),
                        input_processor=GetTimeSeriesAndStaticFeatures(num_time_series_features,window_size),
                        squeeze_output=squeeze_output,
                        use_cuda=use_cuda)
        
        self.agent = SAC(mdp_info, actor_mu_params, actor_sigma_params,
                    actor_optimizer, critic_params, batch_size, initial_replay_size,
                    max_replay_size, warmup_transitions, tau, lr_alpha, log_std_min, log_std_max, target_entropy,
                    critic_fit_params=None)
        
        if agent_name is None:
            self.agent.name = 'SAC_RNN'
        else:
            self.agent.name = agent_name

    def __getattr__(self, attr):
        return getattr(self.agent, attr)

# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/12_SAC_RNN.ipynb 7
class SACRNN_hybrid():

    """
    # TODO Update docstring
    Soft Actor Critic (SAC) agent with hybrid action, both based on Gaussian. The binary action is 
    0 if the output of the network is less or equal than 0, and 1 otherwise.

    Args:
        mdp_info (MDPInfo): Contains relevant information about the environment.
        learning_rate_actor (float): Learning rate for the actor.
        learning_rate_critic (float): Learning rate for the critic.
        learning_rate_alpha (float): Learning rate for the temperature parameter.
        initial_replay_size (int): Number of transitions to save in the replay buffer during startup.
        max_replay_size (int): Maximum number of transitions to save in the replay buffer.
        batch_size (int): Number of transitions to sample each time experience is replayed.
        n_features (int): Number of features for the hidden layers of the networks.
        warmup_transitions (int): Number of transitions to replay before starting to update the networks.
        lr_alpha (float): Learning rate for the temperature parameter.
        tau (float): Parameter for the soft update of the target networks.
        optimizer (torch.optim): Optimizer to use for the networks.
        squeeze_output (bool): Whether to squeeze the output of the actor network or not.
        use_cuda (bool): Whether to use CUDA or not. If True and not available, it will use CPU.
        agent_name (str): Name of the agent. If set to None will use some default name.

    """

    def __init__(
            self,
            mdp_info,
            num_time_series_features,
            num_non_time_series_features,
            window_size,
            learning_rate_actor = 3e-4,
            learning_rate_critic = None,
            initial_replay_size = 64,
            max_replay_size = 50000,
            batch_size = 64,
            n_features = 64,
            warmup_transitions = 100,
            lr_alpha = 3e-4,
            tau = 0.005,
            log_std_min = -20,
            log_std_max = 2,
            target_entropy = None,
            RNN_unit = "GRU", # TODO: Make RNN and LSTM possible
            optimizer = optim.Adam,
            squeeze_output = True,
            use_cuda = True,
            agent_name = None):
        
        use_cuda = use_cuda and torch.cuda.is_available()

        input_shape_data = mdp_info.observation_space.shape
        input_shape_RNN = (num_time_series_features, window_size)
        input_shape_MLP = (num_non_time_series_features,)

        actor_output_shape = (mdp_info.action_space.shape[0],)

        if learning_rate_critic is None:
            learning_rate_critic = learning_rate_actor


        actor_mu_params = dict(network=ActorNetworkRNN_MLP_hybrid,
                                n_features=n_features,
                                input_shape=input_shape_data,
                                input_shape_RNN=input_shape_RNN,
                                input_shape_MLP=input_shape_MLP,
                                output_shape=actor_output_shape,
                                input_processor=GetTimeSeriesAndStaticFeatures(num_time_series_features,window_size),
                                use_cuda=use_cuda)

        actor_sigma_params = dict(network=ActorNetworkRNN_MLP_hybrid,
                                    n_features=n_features,
                                    input_shape= input_shape_data,
                                    input_shape_RNN=input_shape_RNN,
                                    input_shape_MLP=input_shape_MLP,
                                    output_shape=actor_output_shape,
                                    input_processor=GetTimeSeriesAndStaticFeatures(num_time_series_features,window_size),
                                    use_cuda=use_cuda)
        
        actor_optimizer = {'class': optimizer,
                    'params': {'lr': learning_rate_actor}} 
        
        critic_input_shape = (input_shape_data[0] + actor_output_shape[0],)
        input_shape_RNN = (num_time_series_features, window_size)
        print(num_non_time_series_features)
        print(actor_output_shape[0])
        print("XXXXX")
        input_shape_MLP = (num_non_time_series_features + actor_output_shape[0],)

        critic_params = dict(network=CriticNetworkStateActionRNN,
                        optimizer={'class': optim.Adam,
                                'params': {'lr': learning_rate_critic}}, 
                        loss=F.mse_loss,
                        n_features=n_features,
                        input_shape=critic_input_shape,
                        input_shape_RNN = input_shape_RNN,
                        input_shape_MLP = input_shape_MLP,
                        output_shape=(1,),
                        input_processor=GetTimeSeriesAndStaticFeatures(num_time_series_features,window_size),
                        squeeze_output=squeeze_output,
                        use_cuda=use_cuda)
        
        self.agent = SAC(mdp_info, actor_mu_params, actor_sigma_params,
                    actor_optimizer, critic_params, batch_size, initial_replay_size,
                    max_replay_size, warmup_transitions, tau, lr_alpha, log_std_min, log_std_max, target_entropy,
                    critic_fit_params=None)
        
        if agent_name is None:
            self.agent.name = 'SAC_RNN_hybrid'
        else:
            self.agent.name = agent_name

    def __getattr__(self, attr):
        return getattr(self.agent, attr)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/benchmark_agents/11_EOQ.ipynb.

# %% auto 0
__all__ = ['EOQAgent', 'EOQPolicy']

# %% ../../../nbs/agents/benchmark_agents/11_EOQ.ipynb 4
# General libraries:
import numpy as np

# Mushroom libraries
from mushroom_rl.core import Agent

# %% ../../../nbs/agents/benchmark_agents/11_EOQ.ipynb 6
class EOQAgent(Agent):

    train_directly=True
    train_mode = "direct"

    """
    Agent implementing the Economic Order Quantity (EOQ) policy.

    Args:
        mdp_info (MDPInfo): Information about the Markov Decision Process (MDP).
        s (numpy.ndarray): The fixed ordering cost.
        h (numpy.ndarray): The holding cost per unit per period.
        l (numpy.ndarray): The lead time per product.
        preprocessors (list): List of preprocessors to be applied to the state.
        postprocessors (list): List of postprocessors to be applied to the policy.
        agent_name (str): Name of the agent. If set to None will use some default name.
        precision (int): Number of decimal places to round the demand input to.

    Attributes:
        mdp_info (MDPInfo): Information about the Markov Decision Process (MDP).
        policy (EOQPolicy): The EOQ policy implemented by the agent.

    """


    def __init__(self,
                  mdp_info,
                  s, # fixed ordering cost
                  h, # holding cost per unit per period
                  l = None, # lead time
                  preprocessors = None,
                  postprocessors = None,
                  agent_name = None,
                  precision = 5,
                  **kwargs
        ):

        policy = EOQPolicy(
            d = np.zeros(mdp_info.action_space.shape),
            s = s,
            h = h,
            l = l,
            preprocessors = preprocessors,
            postprocessors = postprocessors
        )

        self.precision=precision

        if agent_name is None:
            self.name = 'EOQAgent'
        else:
            self.name = agent_name

        super().__init__(mdp_info, policy)

    def fit(self, features, demand):

        """ 
        Fit the EOQ policy to the given demand.

        This method allows the EOQ agent to adapt its policy to historic demand data, assuming a fixed demand rate without uncertainty.

        Parameters:
            demand (numpy.ndarray): The demand for each period and product. The array should have the shape (num_products, num_periods).

        Returns:
            None

        """


        assert isinstance(demand, np.ndarray)
        assert demand.ndim == 2

        print(demand.shape)
        print(np.round(np.mean(demand, axis=0), self.precision))

        self.policy.d = np.round(np.mean(demand, axis=0), self.precision)

        self.policy.set_q_star()

class EOQPolicy():

    """
    Policy implementing the Economic Order Quantity (EOQ) strategy.

    Notethat d, s, and h must all have the shape (num_products,)

    Args:
        d (numpy.ndarray): The (average) demand per period for each product.
        s (numpy.ndarray): The fixed ordering cost for each product.
        h (numpy.ndarray): The holding cost per unit per period for each product.
        l (numpy.ndarray): The lead time per product.
        postprocessors (list): List of postprocessors to be applied to the action.

    Attributes:
        d (numpy.ndarray): The (average) demand per period for each product.
        s (numpy.ndarray): The fixed ordering cost for each product.
        h (numpy.ndarray): The holding cost per unit per period for each product.
        l (numpy.ndarray): The lead time per product.
        num_products (int): The number of products.
        q_star (numpy.ndarray): The optimal order quantity per product.
        postprocessors (list): List of postprocessors to be applied to the action.

    """

    def __init__(self,
                 d, 
                 s, 
                 h, 
                 l = None,
                 preprocessors = None,
                 postprocessors = None
                 ):
        self.d = d
        self.s = s
        self.h = h
        self.l = l
        self.num_products = len(s)
        self.q_star = 0.5
        if preprocessors is None:
            self.preprocessors = []
        else:
            self.preprocessors = (preprocessors)
        if postprocessors is None:
            self.postprocessors = []
        else:
            self.postprocessors = (postprocessors)

    def set_q_star(self):
        
        """
        Set the optimal order quantity (q_star) for each product.

        This method calculates and assigns the optimal order quantity based on the EOQ formula.

        Returns:
            None

        """

        self.q_star = np.sqrt((2*self.s*self.d)/self.h)

        print("q_star: ", self.q_star)

    def draw_action(self, input):

        """
        Generate an action based on the current state.

        Returns zero for products which have still sufficient inventory, and the optimal order quantity for products which are running out of stock.

        Parameters:
            input (numpy.ndarray): The current inventory level and potentially order pipeline for each product.

        Returns:
            numpy.ndarray: The action to be taken, indicating the quantity to order for each product.

        """

        action = np.zeros(self.num_products)

        for preprocessor in self.preprocessors:
            input = preprocessor(input)

        if self.l is None:
            action = np.where(input >= self.d, 0, self.q_star)

        else:
            pipeline_vector = input[self.num_products:]
            pipeline = np.reshape(pipeline_vector, (self.num_products, max(self.l)))
            pipeline_sum = np.sum(pipeline, axis=1)
            inventory = input[:self.num_products]
            # print("printing inventory, pipeline, lead time, average demand")
            # print(inventory)
            # print(pipeline_sum)
            inventory_position = inventory + pipeline_sum
            # print(self.l)
            # print(self.d)
            action = np.where((inventory_position-self.l*self.d >= 0), 0, self.q_star)
            # print("action: ", action)

        for postprocessor in self.postprocessors:
            action = postprocessor(action)

        return action
    
    def reset(self):
        pass
